/*  xmms-alsa - ALSA output plugin for XMMS
 *  Copyright (C) 1999  Fumihiko Murata.
 *
 *  Contains code Copyright (c) 1998-1999
 *    Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies,
 *    Martin Dahl,Anders Hermansen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "ALSA.h"
//#include <xmms/util.h>
#include <xmms/plugin.h>
#include <errno.h>


#define NFRAGS		32

#define min(x,y) ((x)<(y)?(x):(y))

static snd_pcm_t *handle=NULL;
static snd_pcm_format_t alsa_format;
static gpointer buffer;
static gboolean	going=FALSE,prebuffer,paused=FALSE,unpause=FALSE,do_pause=FALSE,remove_prebuffer=FALSE;
static gboolean write_waiting = FALSE, buffer_waiting = FALSE;
static gint	buffer_size,prebuffer_size,blk_size;
static gint	rd_index=0,wr_index=0;
static gint	output_time_offset=0,written=0,output_bytes=0;
static gint played_bytes=0,pre_pbytes=0;
static gint	bps,ebps,wbps;
static gint	flush;
static gint	fragsize,channels,frequency,efrequency,device_buffer_size;
static gchar	*device_name;
static pthread_t buffer_thread;
static pthread_mutex_t flush_mutex = PTHREAD_MUTEX_INITIALIZER, buffer_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t free_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t flush_cond = PTHREAD_COND_INITIALIZER, buffer_cond = PTHREAD_COND_INITIALIZER;
static pthread_cond_t free_cond = PTHREAD_COND_INITIALIZER;
static gboolean realtime = FALSE;

gint alsa_get_written_time(void)
{
	if(!going)
		return 0;
	return (gint)(((gfloat)written*1000)/(gfloat)(bps));
}

gint alsa_get_output_time(void)
{
	snd_pcm_playback_status_t ps;
	gint bytes,err;
	
	if(!handle||!going)
		return 0;
	
	if(!paused) {
		if(!(err=snd_pcm_playback_status(handle,&ps)))
			bytes = ps.scount+played_bytes;
		else {
			bytes = output_bytes;
			printf("%s\n",snd_strerror(err));
		}
		/*
		printf("sc %8d all %8d pre %8d o %8d pb %8d\r",ps.scount,bytes,
			   pre_pbytes,output_bytes,played_bytes);
		fflush(stdout);
		*/
		/*
		if( bytes < pre_pbytes ) {
			printf("hoe b %d pb %d\n",bytes,pre_pbytes);
		}
		*/
	} else bytes = output_bytes;
	
	if(bytes<0)
		bytes=0;
	pre_pbytes=bytes;
	return output_time_offset+(gint)((gfloat)(bytes*1000.0)/(gfloat)ebps);
}

gint alsa_used(void)
{
	if(realtime)
		return 0;
	else
	{
		if(wr_index>=rd_index)
			return wr_index-rd_index;
		return buffer_size-(rd_index-wr_index);
	}
}


gint alsa_playing(void)
{
	snd_pcm_playback_status_t ps;
	
	snd_pcm_playback_status(handle,&ps);

	if(!alsa_used()&&ps.queue==0)
		return FALSE;
	
	return TRUE;
}


gint alsa_free(void)
{
	if(!realtime)
	{
		if(remove_prebuffer&&prebuffer)
		{
			prebuffer=FALSE;
			remove_prebuffer=FALSE;
		}
		if(prebuffer)
			remove_prebuffer=TRUE;

		if(rd_index>wr_index)
			return (rd_index-wr_index)-device_buffer_size-1;
		return (buffer_size-(wr_index-rd_index))-device_buffer_size-1;
	}
	else
		if ( paused)
			return 0;
		else
			return 1000000;
}

gint alsa_downsample(guchar *ob,guint length,guint speed,guint espeed)
{
	guint nlen,i,off,d, w;
	
	if((alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE)&&channels==2) {
		gulong *nbuffer,*obuffer,*ptr;
	
		obuffer=(gulong *)ob;
		length>>=2;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen<<2);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = snd_pcm_write(handle,nbuffer,nlen<<2);
		g_free(nbuffer);
	}
	else if(((alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE)&&channels==1)
			|| ((alsa_format.format==SND_PCM_SFMT_U8||alsa_format.format==SND_PCM_SFMT_S8)&&channels==2)) {
		gushort *nbuffer,*obuffer,*ptr;
	
		obuffer=(gushort *)ob;
		length>>=1;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen<<1);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = snd_pcm_write(handle,nbuffer,nlen<<1);
		g_free(nbuffer);
	}
	else
	{
		guchar *nbuffer,*obuffer,*ptr;
	
		obuffer=ob;
		
		nlen=(length*espeed)/speed;
		d=(speed<<8)/espeed;

		nbuffer=g_malloc(nlen);
		for(i=0,off=0,ptr=nbuffer;i<nlen;i++)
		{
			*ptr++=obuffer[off>>8];
			off+=d;
		}
		w = snd_pcm_write(handle,nbuffer,nlen);
		g_free(nbuffer);
	}
	return w;

}

void alsa_write(void *ptr,gint length)
{
	gint cnt,off=0, w;
	
	if(!realtime) {
		while (alsa_free() < length )
			xmms_usleep(10000);

		remove_prebuffer = FALSE;
		written+=length;
		while(length>0)
		{
			cnt=min(length,buffer_size-wr_index);
			memcpy(buffer+wr_index,ptr+off,cnt);
			wr_index=(wr_index+cnt)%buffer_size;
			length-=cnt;
			off=cnt;

		}
	}
	else
	{
		if(paused)
			return;

		if(frequency==efrequency)
			w = snd_pcm_write(handle,ptr,length);
		else
			w = alsa_downsample(ptr,length,frequency,efrequency);
		
		if(w == -1 && errno == EIO)
		{
			snd_pcm_close(handle);
			snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,SND_PCM_OPEN_PLAYBACK);
			alsa_set_audio_params();
			if(frequency==efrequency)
				w = snd_pcm_write(handle,ptr,length);
			else
				w = alsa_downsample(ptr,length,frequency,efrequency);
		}
		written += length;
		output_bytes += w;
	}
	
}
	
void alsa_close(void)
{
	wr_index = 0;
	rd_index = 0;
	going = 0;
	if( !realtime)
		pthread_join(buffer_thread, NULL);
	else
	{
		snd_pcm_drain_playback(handle);
		snd_pcm_close(handle);
	}
}

void alsa_flush(gint time)
{
	if(!realtime)
	{
		flush = time;
		while(flush != -1) 
			xmms_usleep(10000);
	}
	else
	{
		snd_pcm_drain_playback(handle);
		/*
		snd_pcm_close(handle);
		snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device);
		alsa_set_audio_params();
		*/
		output_time_offset = time;
		written=(time / 10) * (bps / 100);
		output_bytes=0;
	}
}	

void alsa_pause(short p)
{	
	if(!realtime)
	{
		if(p==TRUE)
			do_pause=TRUE;
		else
			unpause=TRUE;
	}
	else
		paused = p;
}




void *alsa_loop(void *arg)
{
	gint length,cnt, w;
	snd_pcm_playback_status_t ps;

	while(going) {
		if(alsa_used() > prebuffer_size) 
			prebuffer = FALSE;
		if(alsa_used() > 0 && !paused && !prebuffer) {
			length=min(blk_size,alsa_used());
			while(length>0) {
				cnt=min(length,buffer_size-rd_index);
				snd_pcm_playback_status(handle,&ps);
				if( ps.scount == 0 ) {
					//printf("\npb %d to ",played_bytes);
					played_bytes=output_bytes;
					//printf("%d\n",played_bytes);
				}
				if(frequency==efrequency)
					w = snd_pcm_write(handle,buffer+rd_index,cnt);
				else
					w = alsa_downsample(buffer+rd_index,cnt,frequency,efrequency);
				if( w == -1 && errno == EIO) {
					snd_pcm_close(handle);
					snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,
								 SND_PCM_OPEN_PLAYBACK);
					alsa_set_audio_params();
					if(frequency==efrequency)
						w = snd_pcm_write(handle,buffer+rd_index,cnt);
					else
						w = alsa_downsample(buffer+rd_index,cnt,frequency,efrequency);
				}
				output_bytes += w;
				rd_index = (rd_index+cnt) % buffer_size;
				length-=cnt;
			}
			if(!alsa_used())
			  snd_pcm_flush_playback(handle);
		}
		else
			xmms_usleep(10000);
		if(do_pause && !paused) {
			do_pause = FALSE;
			paused = TRUE;
			if( !snd_pcm_playback_status(handle,&ps) ) {
				//printf("\np queue %d %d o %d ",ps.queue,ps.queue&wbps,output_bytes);
				//rd_index -= (abuf_info.fragstotal-abuf_info.fragments) * abuf_info.fragsize;
				rd_index -= output_bytes-(played_bytes+ps.scount) & wbps;
				//rd_index-=ps.queue&wbps;
				//output_bytes -= (abuf_info.fragstotal-abuf_info.fragments) * abuf_info.fragsize;
				output_bytes = (played_bytes+ps.scount) & wbps;
				//output_bytes -=ps.queue&wbps;
				//printf(" o2 %d\n",output_bytes);
			}
			if(rd_index < 0) 
				rd_index += buffer_size;
			snd_pcm_drain_playback(handle);
		}
		if(unpause && paused) {
			unpause = FALSE;
			/*
			snd_pcm_close(handle);
			snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,SND_PCM_OPEN_PLAYBACK);
			alsa_set_audio_params();
			*/
			paused = FALSE;
		}

		if(flush != -1) {
			/*
			 * This close and open is a work around of a bug that exists in some drivers which 
			 * cause the driver to get fucked up by a reset
			 */
			
			snd_pcm_drain_playback(handle);
			/*
			snd_pcm_close(handle);
			snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,SND_PCM_OPEN_PLAYBACK);
			alsa_set_audio_params();
			*/
			output_time_offset=flush;
			written=(flush/10)*(bps/100);
			rd_index = wr_index = output_bytes = 0;
			flush=-1;
			prebuffer = TRUE;
		}			
			
		
	}
	
	snd_pcm_drain_playback(handle);
	snd_pcm_close(handle);
	munlock(buffer);
	g_free(buffer);
	pthread_exit(NULL);
}

void alsa_set_audio_params(void)
{
	snd_pcm_playback_params_t pp;
	snd_pcm_playback_info_t pi;
	gint frag,stereo,err;

	snd_pcm_drain_playback(handle);
	snd_pcm_playback_info(handle,&pi);
	memset(&pp, 0, sizeof(pp));
	pp.fragment_size=1<<fragsize;
	if(pp.fragment_size > pi.max_fragment_size) pp.fragment_size=pi.max_fragment_size;
	if(pp.fragment_size < pi.min_fragment_size) pp.fragment_size=pi.min_fragment_size;
	pp.fragments_max=-1;
	pp.fragments_room=1;
	blk_size=pi.buffer_size;
	if((err=snd_pcm_playback_params(handle,&pp))<0 )
		printf("param err %d\n",snd_strerror(err));
	if((err=snd_pcm_playback_format(handle,&alsa_format))<0 )
		printf("format err %d\n",snd_strerror(err));
	
	efrequency=frequency;
	ebps=efrequency*channels;
	if(alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE)
		ebps*=2;
}



gint alsa_open(AFormat fmt,gint rate,gint nch)
{
	gint fmts,policy;

	memset(&alsa_format, 0, sizeof(alsa_format));
	switch(fmt)
	{
		case FMT_U8:
			alsa_format.format=SND_PCM_SFMT_U8;
			break;
		case FMT_S8:
			alsa_format.format=SND_PCM_SFMT_S8;
			break;
		case FMT_U16_NE:
		case FMT_U16_LE:
			alsa_format.format=SND_PCM_SFMT_U16_LE;
			break;
		case FMT_U16_BE:
			alsa_format.format=-1;
			break;
		case FMT_S16_NE:
		case FMT_S16_LE:
			alsa_format.format=SND_PCM_SFMT_S16_LE;
			break;
		case FMT_S16_BE:
			alsa_format.format=-1;
			break;
	}	
	wbps = -nch;
	bps = rate*nch;
	if(alsa_format.format==SND_PCM_SFMT_S16_LE||alsa_format.format==SND_PCM_SFMT_U16_LE) {
		bps  *= 2;
		wbps *= 2;
	}
	fragsize = 0;
	while((1L << fragsize) < bps / 25)
		fragsize++;
	fragsize--;
	
	device_buffer_size = ((1L << fragsize) * (NFRAGS + 1));

	channels = nch;
	alsa_format.channels = nch;
	frequency = rate;
	alsa_format.rate = rate;
	buffer_size = (alsa_cfg.buffer_size * bps) / 1000;
	if(buffer_size < 8192)
		buffer_size = 8192;
	prebuffer_size = (buffer_size * alsa_cfg.prebuffer) / 100;
	if(buffer_size-prebuffer_size < 4096)
		prebuffer_size = buffer_size - 4096;
	
	buffer_size += device_buffer_size;
	buffer = g_malloc0(buffer_size);
	mlock(buffer, buffer_size);
	
	going = 1;
	flush = -1;
	prebuffer = 1;
	wr_index = rd_index = output_time_offset = written = output_bytes = 0;
	played_bytes=pre_pbytes=0;
	paused = FALSE;
	do_pause = FALSE;
	unpause = FALSE;
	remove_prebuffer = FALSE;

	realtime = xmms_check_realtime_priority();

	if( snd_pcm_open(&handle,alsa_cfg.card,alsa_cfg.pcm_device,SND_PCM_OPEN_PLAYBACK)< 0 ) {
		g_free(buffer);
		return 0;
	}
	alsa_set_audio_params();
	/*
	printf("open buffersize %d prebuffersize %d\n",buffer_size,
		   prebuffer_size);
	*/
	if(!realtime)
		pthread_create(&buffer_thread,NULL,alsa_loop,NULL);
	return 1;
}
