/*
 *  Another ALSA Output plugin 0.3 for xmms
 *  Copyright (c) by Fumihiko Murata <fmurata@p1.tcnet.ne.jp
 *
 *  Contains code Copyright (c) 1998-1999
 *    Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies,
 *    Martin Dahl, Anders Hermansen
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "ALSA.h"

static GtkWidget *configure_win=NULL,*notebook;
static GtkWidget *mainbox,*devicebox,*bbox,*buffer_box;
static GtkWidget *buf_frame,*buf_box,*buf_tbl;
static GtkWidget *buf_siz_box,*buf_siz_label,*buf_siz,*buf_siz_spin;
static GtkWidget *pbuf_siz_box,*pbuf_siz_label,*pbuf_siz,*pbuf_siz_spin;
static GtkObject *buf_siz_adj,*pbuf_siz_adj;
static GtkWidget *card_frame,*card_fbox,*card_box,*card_option_menu;
static GtkWidget *pcm_label,*pcm_box,*pcm_option_menu;
static GtkWidget *mixer_label,*mixer_box,*mixer_option_menu;
static GtkWidget *mixerchannel_label,*mixerchannel_box,*mixerchannel_option_menu;
static GtkWidget *ok,*cancel;
static GtkWidget *card_menu;

typedef struct pcminfoStruct {
	snd_pcm_info_t info;
} pcminfo;

typedef struct mixerchannelinfoStruct {
	snd_mixer_eid_t eid;
	gchar lname[24];
} mixerchannelinfo;

typedef struct mixerinfoStruct {
	snd_mixer_info_t info;
	int e_nums;
	int pcm_no;
	mixerchannelinfo *channel;
} mixerinfo;

typedef struct cardinfoStruct {
	snd_ctl_hw_info_t info;
	pcminfo *pcm;
	mixerinfo *mixer;
} cardinfo;

static cardinfo *cards;

static void pcm_menu(cardinfo *);
static void mixer_menu(cardinfo *);
static void mixerchannel_menu(mixerinfo *);
static void card_callback(GtkWidget *, gint );
static void pcm_callback(GtkWidget *, gint );
static void mixer_callback(GtkWidget *, gint );
static void mixerchannel_callback(GtkWidget *,gint );
static void prep(void);


void configure_win_ok(GtkWidget *widget, gpointer data)
{
	ConfigFile *cfgfile;
	gchar *filename;
	snd_mixer_t *mixer_handle;
	int err;

	alsa_cfg.buffer_size=(gint)GTK_ADJUSTMENT(buf_siz_adj)->value;
	alsa_cfg.prebuffer=(gint)GTK_ADJUSTMENT(pbuf_siz_adj)->value;

	filename=g_strconcat(g_get_home_dir(),"/.xmms/config",NULL);
	if(!(cfgfile=xmms_cfg_open_file(filename)))
		cfgfile=xmms_cfg_new();

	xmms_cfg_write_int(cfgfile,"A-ALSA","card",alsa_cfg.card);
	xmms_cfg_write_int(cfgfile,"A-ALSA","pcm_device",alsa_cfg.pcm_device);
	xmms_cfg_write_int(cfgfile,"A-ALSA","mixer_device",alsa_cfg.mixer_device);
	xmms_cfg_write_string(cfgfile,"A-ALSA","eid_name",alsa_cfg.eid.name);
	xmms_cfg_write_int(cfgfile,"A-ALSA","eid_index",alsa_cfg.eid.index);
	xmms_cfg_write_int(cfgfile,"A-ALSA","buffer_size",alsa_cfg.buffer_size);
	xmms_cfg_write_int(cfgfile,"A-ALSA","prebuffer",alsa_cfg.prebuffer);
	xmms_cfg_write_file(cfgfile,filename);
	xmms_cfg_free(cfgfile);
	g_free(filename);
	gtk_widget_destroy(configure_win);
	snd_mixer_element_free(&alsa_cfg.mixer_element);
	snd_mixer_element_info_free(&alsa_cfg.mixer_info);
	bzero(&alsa_cfg.mixer_element,sizeof(alsa_cfg.mixer_element));
	bzero(&alsa_cfg.mixer_info,sizeof(alsa_cfg.mixer_info));
	alsa_cfg.mixer_element.eid=alsa_cfg.eid;
	alsa_cfg.mixer_info.eid=alsa_cfg.eid;
	if( snd_mixer_open(&mixer_handle,alsa_cfg.card,alsa_cfg.mixer_device)>= 0 )
	{
		snd_mixer_element_build(mixer_handle,&alsa_cfg.mixer_element);
		snd_mixer_element_read(mixer_handle,&alsa_cfg.mixer_element);
		snd_mixer_element_info_build(mixer_handle,&alsa_cfg.mixer_info);
		snd_mixer_element_info(mixer_handle,&alsa_cfg.mixer_info);
		snd_mixer_close(mixer_handle);
		alsa_cfg.mixer_range=alsa_cfg.mixer_info.data.volume1.prange[0].max-
			alsa_cfg.mixer_info.data.volume1.prange[0].min+1;
	}
}

static void pcm_menu(cardinfo *p)
{
	int i;
	GtkWidget *menu,*item;

	menu=gtk_menu_new();
	for( i=0 ; i<p->info.pcmdevs ; i++ ) {
		item=gtk_menu_item_new_with_label(p->pcm[i].info.name);
		gtk_signal_connect(GTK_OBJECT(item),"activate", 
						   GTK_SIGNAL_FUNC(pcm_callback),(gpointer)i);
		gtk_widget_show(item);
		gtk_menu_append(GTK_MENU(menu),item);
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(pcm_option_menu),menu);
}

static void mixer_menu(cardinfo *p)
{
	int i;
	GtkWidget *menu,*item;

	menu=gtk_menu_new();
	for( i=0 ; i<p->info.mixerdevs ; i++ ) {
		item=gtk_menu_item_new_with_label(p->mixer[i].info.name);
		gtk_signal_connect(GTK_OBJECT(item),"activate", 
						   GTK_SIGNAL_FUNC(mixer_callback),(gpointer)i);
		gtk_widget_show(item);
		gtk_menu_append(GTK_MENU(menu),item);
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(mixer_option_menu),menu);
}

static void mixerchannel_menu(mixerinfo *p)
{
	int i;
	GtkWidget *menu,*item;

	menu=gtk_menu_new();
	for( i=0 ; i<p->e_nums ; i++ ) {
		item=gtk_menu_item_new_with_label(p->channel[i].lname);
		gtk_signal_connect(GTK_OBJECT(item),"activate", 
						   GTK_SIGNAL_FUNC(mixerchannel_callback),(gpointer)i);
		gtk_widget_show(item);
		gtk_menu_append(GTK_MENU(menu),item);
	}
	gtk_option_menu_set_menu(GTK_OPTION_MENU(mixerchannel_option_menu),menu);
}

static void card_callback(GtkWidget *widget, gint i)
{
	alsa_cfg.card=i;

	pcm_menu(cards+i);
	alsa_cfg.pcm_device=0;
	mixer_menu(cards+i);
	alsa_cfg.mixer_device=0;
	mixerchannel_menu(cards[i].mixer);
	alsa_cfg.eid_no=cards[i].mixer[0].pcm_no;
	gtk_option_menu_set_history(GTK_OPTION_MENU(mixerchannel_option_menu),
								alsa_cfg.eid_no);
	alsa_cfg.eid=cards[i].mixer[0].channel[alsa_cfg.eid_no].eid;
}

static void pcm_callback(GtkWidget *widget, gint data)
{
	alsa_cfg.pcm_device=data;
}

static void mixer_callback(GtkWidget *widget, gint data)
{
	alsa_cfg.mixer_device=data;
	mixerchannel_menu(cards[alsa_cfg.card].mixer+data);
	alsa_cfg.eid_no=cards[alsa_cfg.card].mixer[data].pcm_no;
	gtk_option_menu_set_history(GTK_OPTION_MENU(mixerchannel_option_menu),
								alsa_cfg.eid_no);
	alsa_cfg.eid=cards[alsa_cfg.card].mixer[data].channel[alsa_cfg.eid_no].eid;

	return;
}

static void mixerchannel_callback(GtkWidget *widget,gint data)
{
	alsa_cfg.eid_no=data;
	alsa_cfg.eid=cards[alsa_cfg.card].mixer[alsa_cfg.mixer_device].channel[data].eid;
	//printf("eid name '%s' index %d\n",alsa_cfg.eid.name,alsa_cfg.eid.index);
}

static void prep(void)
{
	int err,i,j,k,l;
	snd_ctl_t *prep_handle;
	snd_mixer_t *prep_mixer_handle;
	snd_mixer_elements_t elements;
	snd_mixer_eid_t *e_eid;
	unsigned char s[256];
	GtkWidget *item;

	card_menu=gtk_menu_new();
	cards=(cardinfo *)malloc(sizeof(cardinfo)*snd_cards());
	if( !cards ) {
		show_err("Not enough memory");
		return;
	}
	for(i=0;i<snd_cards();i++)
	{
		if((err=snd_ctl_open(&prep_handle,i))<0)
		{
			show_err("open failed");
			printf("open failed: %s\n", snd_strerror(err));
			return;
		}
		if((err=snd_ctl_hw_info(prep_handle, &cards[i].info))<0)
		{
			show_err("hw info failed");
			printf("hw info failed: %s\n", snd_strerror(err));
			snd_ctl_close(prep_handle);
			return;
		}

		//printf("Found soundcard: %s\n", cards[i].info.name);
		item=gtk_menu_item_new_with_label(cards[i].info.name);
		gtk_signal_connect(GTK_OBJECT(item),"activate",
						   GTK_SIGNAL_FUNC(card_callback),(gpointer)i);
		gtk_widget_show(item);
		gtk_menu_append(GTK_MENU(card_menu), item);

		cards[i].pcm = (pcminfo *)malloc(sizeof(pcminfo)*cards[i].info.pcmdevs);
		if( !cards[i].pcm ) {
			show_err("Not enough memory");
			return;
		}
		for(j=0;j<cards[i].info.pcmdevs;j++)
		{
			if((err=snd_ctl_pcm_info(prep_handle, j, &cards[i].pcm[j].info))<0)
			{
				show_err("pcm info failed");
				printf("pcm info failed: %s\n", snd_strerror(err));
				snd_ctl_close(prep_handle);
				return;
			}
		}

		cards[i].mixer=(mixerinfo *)malloc(sizeof(mixerinfo)*cards[i].info.mixerdevs);
		if( !cards[i].mixer ) {
			show_err("Not enough memory");
			return;
		}
		for(j=0;j<cards[i].info.mixerdevs;j++)
		{
			if((err=snd_ctl_mixer_info(prep_handle, j, &cards[i].mixer[j].info))<0)
			{
				show_err("mixer info failed");
				printf("mixer info failed: %s\n", snd_strerror(err));
				snd_ctl_close(prep_handle);
				return;
			}
			if((err=snd_mixer_open(&prep_mixer_handle,i,j))<0)
			{
				show_err("mixer open failed");
				printf("mixer open failed: %s\n", snd_strerror(err));
				return;
			}
			bzero(&elements,sizeof(elements));
			if((err=snd_mixer_elements(prep_mixer_handle,&elements))<0 ) {
				show_err("mixer elements read failed");
				printf("mixer elements read failed 1.\n");
				snd_mixer_close(prep_mixer_handle);
				return;
			}
			elements.pelements=(snd_mixer_eid_t *)
				malloc(elements.elements_over*sizeof(snd_mixer_eid_t));
			if( elements.pelements == NULL ) {
				show_err("Not enough memory.");
				return;
			}
			elements.elements_size = elements.elements_over;
			elements.elements_over = elements.elements = 0;
			//printf("elements %d\n",elements.elements_size);
			if((err=snd_mixer_elements(prep_mixer_handle,&elements))<0 ) {
				show_err("mixer elements read failed");
				printf("mixer elements read failed 2.\n");
				snd_mixer_close(prep_mixer_handle);
				return;
			}
			l=0;
			for(k=0;k<elements.elements;k++ )
				if( elements.pelements[k].type == SND_MIXER_ETYPE_VOLUME1 ) l++;
			cards[i].mixer[j].e_nums=l;
			e_eid=elements.pelements;
			cards[i].mixer[j].channel = (mixerchannelinfo *)
				malloc(sizeof(mixerchannelinfo)*l);
			if( !cards[i].mixer[j].channel ) {
				show_err("Not enough memory");
				return;
			}
			cards[i].mixer[j].pcm_no=0;
			for(k=0;k<l;k++)
			{
				while( e_eid->type != SND_MIXER_ETYPE_VOLUME1 ) e_eid++;
				cards[i].mixer[j].channel[k].eid=*e_eid;
				if( (e_eid->name[0]|0x20) == 'p' &&
					(e_eid->name[1]|0x20) == 'c' &&
					(e_eid->name[2]|0x20) == 'm' ) cards[i].mixer[j].pcm_no=k;
				if( cards[i].mixer[j].channel[k].eid.index > 0 )
					sprintf(cards[i].mixer[j].channel[k].lname,"%s %d",
							cards[i].mixer[j].channel[k].eid.name,
							cards[i].mixer[j].channel[k].eid.index);
				else
					sprintf(cards[i].mixer[j].channel[k].lname,"%s",
							cards[i].mixer[j].channel[k].eid.name);
				e_eid++;
			}
			free(elements.pelements);
			snd_mixer_close(prep_mixer_handle);
		}
		snd_ctl_close(prep_handle);
	}

	return;
}

void alsa_configure(void)
{
	int i;

	if(!configure_win)
	{
		// Mainwindow
		configure_win=gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_signal_connect(GTK_OBJECT(configure_win),"destroy",GTK_SIGNAL_FUNC(gtk_widget_destroyed),&configure_win);
		gtk_window_set_title(GTK_WINDOW(configure_win),"Configure Another ALSA Driver 0.3");
		gtk_container_border_width(GTK_CONTAINER(configure_win),10);

		// Mainbox
		mainbox=gtk_vbox_new(FALSE,10);
		gtk_container_add(GTK_CONTAINER(configure_win), mainbox);

		// Bufferbox
		buffer_box=gtk_vbox_new(FALSE,10);
		gtk_container_border_width(GTK_CONTAINER(buffer_box),10);

		// Devicebox
		devicebox=gtk_vbox_new(FALSE,10);
		gtk_container_border_width(GTK_CONTAINER(devicebox),10);

		// Find sounddevices
		prep();

		// Soundcard
		card_frame=gtk_frame_new("Sound Card:");
		gtk_box_pack_start(GTK_BOX(devicebox),card_frame,FALSE,FALSE,0);

		card_fbox=gtk_vbox_new(FALSE, 0);
		gtk_container_border_width(GTK_CONTAINER(card_fbox),10);
		gtk_container_add(GTK_CONTAINER(card_frame), card_fbox);

		card_option_menu=gtk_option_menu_new();
		gtk_box_pack_start(GTK_BOX(card_fbox), card_option_menu,TRUE,TRUE,0);
		gtk_option_menu_set_menu(GTK_OPTION_MENU(card_option_menu),card_menu);
		gtk_option_menu_set_history(GTK_OPTION_MENU(card_option_menu),alsa_cfg.card);
		card_box=gtk_vbox_new(FALSE, 0);
		gtk_container_border_width(GTK_CONTAINER(card_box),0);
		gtk_container_add(GTK_CONTAINER(card_fbox), card_box);

		gtk_widget_show(card_option_menu);
		gtk_widget_show(card_fbox);
		gtk_widget_show(card_box);
		gtk_widget_show(card_frame);

		// Pcm
		pcm_label=gtk_label_new("PCM device:");
		gtk_box_pack_start(GTK_BOX(card_box),pcm_label,FALSE,FALSE,0);
		gtk_misc_set_alignment (GTK_MISC (pcm_label), 0, 0);

		pcm_option_menu=gtk_option_menu_new();
		gtk_box_pack_start(GTK_BOX(card_box), pcm_option_menu,TRUE,TRUE,0);
		pcm_menu(cards+alsa_cfg.card);
		gtk_option_menu_set_history(GTK_OPTION_MENU(pcm_option_menu),
									alsa_cfg.pcm_device);

		gtk_widget_show(pcm_label);
		gtk_widget_show(pcm_option_menu);

		// Mixer
		mixer_label=gtk_label_new("Mixer device:");
		gtk_box_pack_start(GTK_BOX(card_box),mixer_label,FALSE,TRUE,0);
		gtk_misc_set_alignment (GTK_MISC (mixer_label), 0, 0);

		mixer_box=gtk_vbox_new(FALSE, 0);
		gtk_container_border_width(GTK_CONTAINER(mixer_box),0);
		gtk_container_add(GTK_CONTAINER(card_box), mixer_box);

		mixer_option_menu=gtk_option_menu_new();
		gtk_box_pack_start(GTK_BOX(mixer_box), mixer_option_menu,TRUE,TRUE,0);
		mixer_menu(cards+alsa_cfg.card);
		gtk_option_menu_set_history(GTK_OPTION_MENU(mixer_option_menu),
									alsa_cfg.mixer_device);

		gtk_widget_show(mixer_label);
		gtk_widget_show(mixer_option_menu);
		gtk_widget_show(mixer_box);

		// Mixerchannel
		mixerchannel_label=gtk_label_new("Mixer channel:");
		gtk_box_pack_start(GTK_BOX(mixer_box),mixerchannel_label,FALSE,FALSE,0);
		gtk_misc_set_alignment (GTK_MISC (mixerchannel_label), 0, 0);

		mixerchannel_option_menu=gtk_option_menu_new();
		gtk_box_pack_start(GTK_BOX(mixer_box), mixerchannel_option_menu,TRUE,TRUE,0);
		mixerchannel_menu(cards[alsa_cfg.card].mixer+alsa_cfg.mixer_device);
		for( i = 0 ; i<cards[alsa_cfg.card].mixer[alsa_cfg.mixer_device].e_nums
				 ; i++ ) {
			if(strcmp(alsa_cfg.eid.name,cards[alsa_cfg.card].mixer[alsa_cfg.mixer_device].channel[i].eid.name) == 0 ) {
				alsa_cfg.eid_no=i;
				//i=cards[alsa_cfg.card].mixer[alsa_cfg.mixer_device].e_nums;
				break;
			}
		}
		gtk_option_menu_set_history(GTK_OPTION_MENU(mixerchannel_option_menu),
									alsa_cfg.eid_no);

		gtk_widget_show(mixerchannel_label);
		gtk_widget_show(mixerchannel_option_menu);

		// Buffer frame
		buf_frame=gtk_frame_new("Buffering:");
		gtk_box_pack_start(GTK_BOX(buffer_box),buf_frame,FALSE,FALSE,0);

		buf_box=gtk_vbox_new(FALSE,0);
		gtk_container_border_width(GTK_CONTAINER(buf_box),10);
		gtk_container_add(GTK_CONTAINER(buf_frame),buf_box);

		buf_tbl=gtk_table_new(1,2,TRUE);
		gtk_container_border_width(GTK_CONTAINER(buf_tbl),0);
		gtk_box_pack_start(GTK_BOX(buf_box),buf_tbl,FALSE,FALSE,0);

		buf_siz_box=gtk_hbox_new(FALSE,0);
		gtk_table_attach_defaults(GTK_TABLE(buf_tbl),buf_siz_box,0,1,0,1);
		buf_siz_label=gtk_label_new("Buffer size (ms):");
		gtk_box_pack_start(GTK_BOX(buf_siz_box),buf_siz_label,FALSE,FALSE,0);
		gtk_widget_show(buf_siz_label);
		buf_siz_adj=gtk_adjustment_new(alsa_cfg.buffer_size,200,10000,100,100,100);
		buf_siz_spin=gtk_spin_button_new(GTK_ADJUSTMENT(buf_siz_adj),8,0);
		gtk_widget_set_usize(buf_siz_spin,60,-1);
		gtk_box_pack_start(GTK_BOX(buf_siz_box),buf_siz_spin,FALSE,FALSE,0);
		gtk_widget_show(buf_siz_spin);
		gtk_widget_show(buf_siz_box);

		pbuf_siz_box=gtk_hbox_new(FALSE,0);
		gtk_table_attach_defaults(GTK_TABLE(buf_tbl),pbuf_siz_box,0,1,2,3);
		pbuf_siz_label=gtk_label_new("Pre-buffer (percent):");
		gtk_box_pack_start(GTK_BOX(pbuf_siz_box),pbuf_siz_label,FALSE,FALSE,0);
		gtk_widget_show(pbuf_siz_label);
		pbuf_siz_adj=gtk_adjustment_new(alsa_cfg.prebuffer,0,90,1,1,1);
		pbuf_siz_spin=gtk_spin_button_new(GTK_ADJUSTMENT(pbuf_siz_adj),1,0);
		gtk_widget_set_usize(pbuf_siz_spin,60,-1);
		gtk_box_pack_start(GTK_BOX(pbuf_siz_box),pbuf_siz_spin,FALSE,FALSE,0);
		gtk_widget_show(pbuf_siz_spin);
		gtk_widget_show(pbuf_siz_box);
		gtk_widget_show(buf_tbl);
		gtk_widget_show(buf_box);
		gtk_widget_show(buf_frame);

		// Notebook
		notebook=gtk_notebook_new();
		gtk_box_pack_start(GTK_BOX(mainbox),notebook,TRUE,TRUE,0);
		gtk_notebook_append_page(GTK_NOTEBOOK(notebook),devicebox,gtk_label_new("Devices"));
		gtk_notebook_append_page(GTK_NOTEBOOK(notebook),buffer_box,gtk_label_new("Buffering"));

		// Buttonbox
		bbox=gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox),GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox),5);
		gtk_box_pack_start(GTK_BOX(mainbox),bbox,FALSE,FALSE,0);

		// Buttons
		ok=gtk_button_new_with_label("Ok");
		gtk_signal_connect(GTK_OBJECT(ok),"clicked",GTK_SIGNAL_FUNC(configure_win_ok),NULL);
		GTK_WIDGET_SET_FLAGS(ok,GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox),ok,TRUE,TRUE,0);
		gtk_widget_show(ok);
		gtk_widget_grab_default(ok);

		cancel=gtk_button_new_with_label("Cancel");
		gtk_signal_connect_object(GTK_OBJECT(cancel),"clicked",GTK_SIGNAL_FUNC(gtk_widget_destroy),GTK_OBJECT(configure_win));
		GTK_WIDGET_SET_FLAGS(cancel,GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox),cancel,TRUE,TRUE,0);
		gtk_widget_show(cancel);

		// Show main widgets
		gtk_widget_show(devicebox);
		gtk_widget_show(buffer_box);
		gtk_widget_show(notebook);
		gtk_widget_show(bbox);
		gtk_widget_show(mainbox);
		gtk_widget_show(configure_win);
	}
	else
		gdk_window_raise(configure_win->window);
}

static GtkWidget *error_win=NULL,*mainbox,*err_label;
static GtkWidget *bbox,*ok;

void error_win_ok(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(error_win);
}

void show_err(char *errmsg)
{
	if(!error_win)
	{
		// Error window
		error_win=gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_signal_connect(GTK_OBJECT(error_win),"destroy",GTK_SIGNAL_FUNC(gtk_widget_destroyed),&error_win);
		gtk_window_set_title(GTK_WINDOW(error_win),"Error");
		gtk_container_border_width(GTK_CONTAINER(error_win),10);

		// Mainbox
		mainbox=gtk_vbox_new(FALSE,10);
		gtk_container_add(GTK_CONTAINER(error_win), mainbox);

		err_label=gtk_label_new(errmsg);
		gtk_box_pack_start(GTK_BOX(mainbox),err_label,FALSE,FALSE,0);
		gtk_widget_show(err_label);

		// Buttonbox
		bbox=gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox),GTK_BUTTONBOX_END);
		gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox),5);
		gtk_box_pack_start(GTK_BOX(mainbox),bbox,FALSE,FALSE,0);

		// Buttons
		ok=gtk_button_new_with_label("Ok");
		gtk_signal_connect(GTK_OBJECT(ok),"clicked",GTK_SIGNAL_FUNC(error_win_ok),NULL);
		GTK_WIDGET_SET_FLAGS(ok,GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox),ok,TRUE,TRUE,0);
		gtk_widget_show(ok);
		gtk_widget_grab_default(ok);

		// Show main widgets
		gtk_widget_show(bbox);
		gtk_widget_show(mainbox);
		gtk_widget_show(error_win);
	}
	else
		gdk_window_raise(error_win->window);
}
